
function Create(self)
	---------------- AI variables start ----------------
	self.StableTimer = Timer()
	self.DoorTimer = Timer()
	
	self.ObstacleTimer = Timer()
	self.ObstacleTimer:SetSimTimeLimitMS(100)
	
	self.PlayerInterferedTimer = Timer()
	self.PlayerInterferedTimer:SetSimTimeLimitMS(500)
	
	self.DeliveryState = ACraft.FALL
	self.LastAIMode = Actor.AIMODE_NONE
	self.Controller = self:GetController()
	self.groundDist = self.Radius / 1.35
	
	self.LZpos = SceneMan:MovePointToGround(self.Pos, self.groundDist, 5)
	self.velIntegrator = 0
	
	if self.AIMode == Actor.AIMODE_DELIVER and self:IsInventoryEmpty() then
		self.AIMode = Actor.AIMODE_STAY	-- stop the craft from returning to orbit immediately
	end
	
	self.MoveLZ = function()
		local FuturePos = self.Pos + self.Vel*7
		
		-- Make sure FuturePos is inside the scene
		if FuturePos.X > SceneMan.SceneWidth then
			if SceneMan.SceneWrapsX then
				FuturePos.X = FuturePos.X - SceneMan.SceneWidth
			else
				FuturePos.X = SceneMan.SceneWidth - self.Radius
			end
		elseif FuturePos.X < 0 then
			if SceneMan.SceneWrapsX then
				FuturePos.X = FuturePos.X + SceneMan.SceneWidth
			else
				FuturePos.X = self.Radius
			end
		end
		
		if self.DeliveryState == ACraft.LAUNCH then
			self.LZpos.X = FuturePos.X
		else
			local TestPos = Vector(FuturePos.X, math.min(FuturePos.Y, self.Pos.Y))
			self.LZpos = SceneMan:MovePointToGround(TestPos, self.groundDist, 5)
		end
	end
	
	-- The controller class
	PIDController = {}
	PIDController.mt = {__index = PIDController}
	function PIDController:New(p_in, i_in, d_in, last_in, integral_in)
		return setmetatable(
			{
				p = p_in,
				i = i_in,
				d = d_in,
				last = last_in,
				integral = integral_in or 0,
			},
			PIDController.mt)
	end
	
	function PIDController:Update(input, target)
		local err = input - target
		local change = input - self.last
		self.integral = self.integral + err
		self.last = input
		return self.p*err + self.i*self.integral + self.d*change
	end
	
	function PIDController:UpdateLimit(input, target, max)
		local err = input - target
		local change = input - self.last
		self.integral = math.min(math.max(self.integral + self.i*err, -max), max)
		self.last = input
		return self.p*err + self.integral + self.d*change
	end
	
	-- The controllers
	self.AngPID = PIDController:New(3, 0.005, 1, self.RotAngle)
	self.XposPID = PIDController:New(0.2, 0, 0.5, 0)
	self.YposPID = PIDController:New(0.012, 0.00005, 4, self.LZpos.Y)
	
	---------------- AI variables end ----------------
end

function UpdateAI(self)
	if self.PlayerInterferedTimer:IsPastSimTimeLimit() then
		self.MoveLZ()
	end
	
	self.PlayerInterferedTimer:Reset()
	
	if self.AIMode ~= self.LastAIMode then
		self.LastAIMode = self.AIMode

		if self.AIMode == Actor.AIMODE_RETURN then
			self.DeliveryState = ACraft.LAUNCH
			self.LZpos.Y = -10000	-- Go to orbit
			self.MoveLZ()
		else
			self.DeliveryState = ACraft.FALL
			self.MoveLZ()
		end
	end
	
	-- Reset the timer if not in a stable and upright state
	self.velIntegrator = self.velIntegrator * 0.7 + self.Vel.Magnitude * 0.3
	if self.velIntegrator > 2.5 or math.abs(self.AngularVel) > 1 then
		self.StableTimer:Reset()
	else
		self.LZpos.X = self.Pos.X
	end
	
	-- Delivery Sequence logic
	if self.DeliveryState == ACraft.FALL then		
		self.checkLZ = not self.checkLZ	-- Only check every second frame
		if self.checkLZ then
			-- Move Landing Zone closer to the rocket if it is about to collide with something
			if (self.LZpos.Y-self.Pos.Y) > self.Radius then
				local Speed = (self.Vel + Vector(0,4))*15
				if SceneMan:CastObstacleRay(self.Pos, Speed, Vector(), Vector(), self.ID, 0, 10) > -1 then
					self.MoveLZ()
				end
			end
		else
			-- Check for something in the way of our descent, and move to the side to avoid it
			if self.ObstacleTimer:IsPastSimTimeLimit() then
				self.ObstacleTimer:Reset()
				
				local Trace = Vector(self.Vel.X+RangeRand(-1,1), math.max(self.Vel.Y,1)) * 40
				local obstID = SceneMan:CastMORay(self.Pos, Trace, self.ID, 0, false, 5)
				if obstID < 255 then
					local MO = MovableMan:GetMOFromID(MovableMan:GetRootMOID(obstID))
					if MO.ClassName == "ACDropShip" or MO.ClassName == "ACRocket" then
						self.LZpos:SetXY(MO.Pos.X+MO.Radius*3, math.max(self.Pos.Y-200,0))
						self.ObstacleTimer:SetSimTimeLimitMS(750)
						self.obstacle = true
					end
				elseif self.obstacle then
					self.obstacle = false
					self.MoveLZ()
					self.ObstacleTimer:SetSimTimeLimitMS(100)
				end
			end
		end
		
		if self.AIMode == Actor.AIMODE_DELIVER and self:IsInventoryEmpty() then
			self.DeliveryState = ACraft.LAUNCH	-- Don't descend if we have nothing to deliver
			self.LZpos.Y = -10000	-- Go to orbit
		else
			if self.StableTimer:IsPastSimMS(500) then	-- Move LZ if stable
				self.LZpos = SceneMan:MovePointToGround(Vector(self.Pos.X, self.Pos.Y), self.groundDist, 6)
			end
			
			if self.AIMode ~= Actor.AIMODE_STAY then
				local dist = SceneMan:ShortestDistance(self.Pos, self.LZpos, false).Magnitude
				if dist < 20 then	-- If we passed the check, start unloading
					self.DeliveryState = ACraft.UNLOAD
				end
			end
		end
	elseif self.DeliveryState == ACraft.UNLOAD then
		if self:IsInventoryEmpty() then	-- Return to orbit if empty
			self:CloseHatch()
			
			if self.DoorTimer:IsPastSimMS(2000) then	-- Pause before returning to orbit
				self.DeliveryState = ACraft.LAUNCH
				self.LZpos.Y = -10000	-- Go to orbit
			end
		elseif self.StableTimer:IsPastSimMS(500) then
			self.Controller:SetState(Controller.PRESS_FACEBUTTON, true)	-- Open hatch
			self.DoorTimer:Reset()
		end
	elseif self.DeliveryState == ACraft.LAUNCH then
		-- Check for something in the way of our ascent, and move to the side to avoid it
		if self.ObstacleTimer:IsPastSimTimeLimit() then
			self.ObstacleTimer:Reset()
			
			local Trace = Vector(self.Vel.X+RangeRand(-1,1), math.min(self.Vel.Y,-1)) * 50
			local obstID = SceneMan:CastMORay(self.Pos, Trace, self.ID, 0, false, 5)
			if obstID < 255 then
				local MO = MovableMan:GetMOFromID(MovableMan:GetRootMOID(obstID))
				if MO.ClassName == "ACDropShip" or MO.ClassName == "ACRocket" then
					self.LZpos:SetXY(MO.Pos.X-(MO.Diameter+self.Diameter), self.Pos.Y+self.Vel.Y)
					self.ObstacleTimer:SetSimTimeLimitMS(750)
					self.obstacle = true
				end
			elseif self.obstacle then
				self.obstacle = false
				self.LZpos.Y = -10000	-- Go to orbit
				self.MoveLZ()
				self.ObstacleTimer:SetSimTimeLimitMS(100)
			else
				self.MoveLZ()	-- move the exit point closer towards us
			end
		end
	end
	
	-- Control up/down movement
	if self.DeliveryState ~= ACraft.UNLOAD then
		local limit = self.Mass * -0.026 + 27	-- Descend slower when carrying a heavy cargo
		local change = self.YposPID:UpdateLimit(-(self.LZpos.Y-self.Pos.Y), 0, limit)
		if math.abs(self.RotAngle) < 0.8 then
			if change > 4 then
				self.Controller:SetState(Controller.MOVE_UP, true)
			elseif change < -15 then
				self.Controller:SetState(Controller.MOVE_DOWN, true)
			end
		elseif self.RotAngle > 2.14 and self.RotAngle < 4.14 then	-- Upside down
			self.Controller:SetState(Controller.MOVE_DOWN, true)
		end
	end
	
	-- Control right/left movement (the rocket will move sideways if rotated to the side)
	local dist = SceneMan:ShortestDistance(self.Pos+self.Vel*20, self.LZpos, false).X
	change = self.XposPID:Update(dist, 0)
	local targetAng = 0
	if self.Vel.Y > 0 then
		if change < -4 then
			targetAng = -math.max(change/40, -0.5)
		elseif change > 4 then
			targetAng = -math.min(change/40, 0.5)
		end
	else
		if change > 4 then
			targetAng = -math.max(change/40, -0.5)
		elseif change < -4 then
			targetAng = -math.min(change/40, 0.5)
		end
	end
	
	-- Control angle
	change = self.AngPID:Update(self.RotAngle+self.AngularVel, targetAng)
	if change > 1 and not self.burstRight then
		self.burstRight = math.max(6-change, 2) -- Wait n frames until next burst (lower -> better control)
	elseif change < -1 and not self.burstLeft then
		self.burstLeft = math.max(6+change, 2)
	end
	
	-- Don't fire the side thrusters if they can hit friendly actors
	if self.DeliveryState == ACraft.UNLOAD or self.DeliveryState == ACraft.LAUNCH then
		if self.burstLeft and self.burstLeft < 0 then
			local Start = self.Pos+self:RotateOffset(Vector(-13,-16))
			local MOID = SceneMan:CastMORay(Start, self:RotateOffset(Vector(-20,0)), self.ID, 0, false, 3)
			if MOID < 255 then
				local Act = MovableMan:GetMOFromID(MovableMan:GetRootMOID(MOID))
				if Act and Act.Team == self.Team then
					self.burstLeft = 10	-- Someting is in front of the thruster: don't fire
				end
			end
		end
		
		if self.burstRight and self.burstRight < 0 then
			local Start = self.Pos+self:RotateOffset(Vector(12,-16))
			local MOID = SceneMan:CastMORay(Start, self:RotateOffset(Vector(20,0)), self.ID, 0, false, 3)
			if MOID < 255 then
				local Act = MovableMan:GetMOFromID(MovableMan:GetRootMOID(MOID))
				if Act and Act.Team == self.Team then
					self.burstRight = 10	-- Someting is in front of the thruster: don't fire
				end
			end
		end
	end
	
	if self.DoorTimer.ElapsedRealTimeMS < 800 then
		if self.exitLeft then
			self.burstRight = nil
		else
			self.burstLeft = nil
		end
	end
	
	-- Trigger bursts
	if self.burstRight then
		self.burstRight = self.burstRight - 1
		if self.burstRight < 0 then			
			self.Controller:SetState(Controller.MOVE_RIGHT, true)
			if self.burstRight < -8 then	-- Fire for -n frames (higher -> better control)
				self.burstRight = nil	-- Allow a new burst
			end
		end
	end
	
	if self.burstLeft then
		self.burstLeft = self.burstLeft - 1
		if self.burstLeft < 0 then
			self.Controller:SetState(Controller.MOVE_LEFT, true)
			if self.burstLeft < -8 then
				self.burstLeft = nil
			end
		end
	end
end



function OldUpdateAI(self)	
	-- Reset the timer if not in a stable and upright state
	if self.Vel.Magnitude > 4 or math.abs(self.AngularVel) > 1 then
		self.StableTimer:Reset()
	else
		self.LZpos.X = self.Pos.X
		self:CloseHatch()
	end
	
	if self.AIMode == Actor.AIMODE_RETURN then
		self.DeliveryState = ACraft.LAUNCH
	end

	local range = self.LZpos.Y - self.Pos.Y -- Y distance to LZ
	
	-- Delivery Sequence logic
	if self.DeliveryState == ACraft.FALL then		
		self.checkLZ = not self.checkLZ	-- Only check every second frame
		if self.checkLZ then
			-- Move Landing Zone closer to the rocket if it is about to collide with something
			if range > self.Radius then
				local Speed = (self.Vel + Vector(0,4))*15
				local dist = SceneMan:CastObstacleRay(self.Pos, Speed, Vector(), Vector(), self.ID, 0, 10)
				if dist > -1 then
					local NewPos = self.Pos + SceneMan:ShortestDistance(self.Pos, self.LZpos, false)*0.7
					
					if SceneMan.SceneWrapsX then
						if NewPos.X > SceneMan.SceneWidth then
							NewPos.X = NewPos.X - SceneMan.SceneWidth
						elseif NewPos.X < 0 then
							NewPos.X = NewPos.X + SceneMan.SceneWidth
						end
					end
					
					self.LZpos = SceneMan:MovePointToGround(Vector(NewPos.X, self.Pos.Y), self.groundDist, 9)
				end
			end
		else
			-- Check for something in the way of our descent, and move to the side to avoid it
			if self.ObstacleTimer:IsPastSimMS(self.checkTime) then
				self.ObstacleTimer:Reset()
				
				local Trace = Vector(self.Vel.X+RangeRand(-1,1), math.max(self.Vel.Y,1)) * 40
				local obstID = SceneMan:CastMORay(self.Pos, Trace, self.ID, 0, false, 5)
				if obstID < 255 then
					local MO = MovableMan:GetMOFromID(MovableMan:GetRootMOID(obstID))
					if MO.ClassName == "ACDropShip" or MO.ClassName == "ACRocket" then
						self.LZpos:SetXY(MO.Pos.X+MO.Radius*3, math.max(self.Pos.Y-200,0))
						self.ObstacleTimer:Reset()
						self.checkTime = 500
					end
				else
					self.LZpos = SceneMan:MovePointToGround(self.Pos, self.groundDist, 9)
					self.checkTime = 100
				end
			end
		end
		
		if self.AIMode ~= Actor.AIMODE_STAY then
			-- Don't descend if we have nothing to deliver
			if self:IsInventoryEmpty() then
				self.DeliveryState = ACraft.LAUNCH
			else
				if self.StableTimer:IsPastSimMS(400) then	-- Move LZ if stable
					self.LZpos = SceneMan:MovePointToGround(Vector(self.Pos.X, self.Pos.Y), self.groundDist, 6)
				end
				
				local dist = SceneMan:ShortestDistance(self.Pos, self.LZpos, false).Magnitude
				-- If we passed the check, start unloading
				if dist < 15 then
					self.DeliveryState = ACraft.UNLOAD
				end
			end
		end
	elseif self.DeliveryState == ACraft.UNLOAD then
		if self:IsInventoryEmpty() then	-- Return to orbit if empty
			if self.DoorTimer:IsPastSimMS(2000) then	-- Pause before returning to orbit
				self.DeliveryState = ACraft.LAUNCH
				self:CloseHatch()
			end
		elseif self.StableTimer:IsPastSimMS(500) then
			self.Controller:SetState(Controller.PRESS_FACEBUTTON, true)	-- Open hatch
			self.DoorTimer:Reset()
		end
	elseif self.DeliveryState == ACraft.LAUNCH then
		-- Check for something in the way of our ascent, and move to the side to avoid it
		if self.ObstacleTimer:IsPastSimMS(self.checkTime) then
			self.ObstacleTimer:Reset()
			
			local Trace = Vector(self.Vel.X+RangeRand(-1,1), math.min(self.Vel.Y,-1)) * 35
			local obstID = SceneMan:CastMORay(self.Pos, Trace, self.ID, 0, false, 5)
			if obstID < 255 then
				local MO = MovableMan:GetMOFromID(MovableMan:GetRootMOID(obstID))
				if MO.ClassName == "ACDropShip" or MO.ClassName == "ACRocket" then
					self.LZpos:SetXY(MO.Pos.X-MO.Radius*3, self.Pos.Y-5)
					self.checkTime = 500
				end
			else
				self.LZpos:SetXY(self.Pos.X, -10000)	-- Go to orbit
				self.checkTime = 100
			end
		end
	end
	
	-- Control up/down movement
	if self.DeliveryState ~= ACraft.UNLOAD then
		local acc = self.LastVel.Y - self.Vel.Y
		local limit = self.Mass * -0.026 + 27	-- Descend slower when carrying a heavy cargo
		local change = self.YposPID:UpdateLimit(-range, 0, limit)
		if math.abs(self.RotAngle) < 0.8 then
			if change > 4 then
				self.Controller:SetState(Controller.MOVE_UP, true)
			else
				self.Controller:SetState(Controller.MOVE_UP, false)
				
				if change < -15 then
					self.burstUp = nil
					self.Controller:SetState(Controller.MOVE_DOWN, true)
				end
			end
		else
			self.Controller:SetState(Controller.MOVE_UP, false)
			
			if self.RotAngle > 2.14 and self.RotAngle < 4.14 then	-- Upside down
				self.Controller:SetState(Controller.MOVE_DOWN, true)
			end
		end
	end
	
	-- Control right/left movement (the rocket will move sideways if rotated to the side)
	local dist = SceneMan:ShortestDistance(self.Pos+self.Vel*20, self.LZpos, false).X
	change = self.XposPID:Update(dist, 0)
	local targetAng = 0
	if self.Vel.Y > -1 then
		if self.DeliveryState == ACraft.UNLOAD then
			if change < -6 then
				targetAng = -math.max(change/40, -0.6)
			elseif change > 6 then
				targetAng = -math.min(change/40, 0.6)
			end
		else
			if change < -2 then
				targetAng = -math.max(change/40, -0.6)
			elseif change > 2 then
				targetAng = -math.min(change/40, 0.6)
			end
		end
	else
		if change > 2 then
			targetAng = -math.max(change/40, -0.7)
		elseif change < -2 then
			targetAng = -math.min(change/40, 0.7)
		end
	end
	
	-- Control angle
	change = self.AngPID:Update(self.RotAngle+self.AngularVel, targetAng)
	if change > 1 and not self.burstRight then
		self.burstRight = math.max(7-change, 1) -- Wait n frames until next burst (lower -> better control)
	elseif change < -1 and not self.burstLeft then
		self.burstLeft = math.max(7+change, 1)
	end
	
	-- Don't fire the side thrusters when the doors are open
	if self.DoorTimer.ElapsedRealTimeMS < 800 then
		if self.exitLeft then
			self.burstRight = nil
		else
			self.burstLeft = nil
		end
	end
	
	-- Trigger bursts
	if self.burstRight then
		self.burstRight = self.burstRight - 1
		if self.burstRight < 0 then
			self.Controller:SetState(Controller.MOVE_RIGHT, true)
			if self.burstRight < -3 then	-- Fire for -m frames (higher -> better control)
				self.burstRight = nil	-- Allow a new burst
			end
		else
			self.Controller:SetState(Controller.MOVE_RIGHT, false)
		end
	else
		self.Controller:SetState(Controller.MOVE_RIGHT, false)
	end
	
	if self.burstLeft then
		self.burstLeft = self.burstLeft - 1
		if self.burstLeft < 0 then
			self.Controller:SetState(Controller.MOVE_LEFT, true)
			if self.burstLeft < -3 then
				self.burstLeft = nil
			end
		else
			self.Controller:SetState(Controller.MOVE_LEFT, false)
		end
	else
		self.Controller:SetState(Controller.MOVE_LEFT, false)
	end
	
	self.LastVel:SetXY(self.Vel.X, self.Vel.Y)
end
